@@ -1,6 +1,6 @@
 package HTML::Template;
 
-$HTML::Template::VERSION = '1.0';
+$HTML::Template::VERSION = '1.2.1';
 
 =head1 NAME
 
@@ -31,8 +31,10 @@ Now create a small CGI program:
   my $template = HTML::Template->new(filename => 'test.tmpl');
 
   # fill in some parameters
-  $template->param('HOME', $ENV{HOME});
-  $template->param('PATH', $ENV{PATH});
+  $template->param(
+      HOME => $ENV{HOME},
+      PATH => $ENV{PATH},
+  );
 
   # send the obligatory Content-Type
   print "Content-Type: text/html\n\n";
@@ -48,17 +50,17 @@ My Path is set to /bin;/usr/bin
 
 =head1 DESCRIPTION
 
-This module attempts make using HTML templates simple and natural.  It
+This module attempts to make using HTML templates simple and natural.  It
 extends standard HTML with a few new HTML-esque tags - <TMPL_VAR>,
 <TMPL_LOOP>, <TMPL_INCLUDE>, <TMPL_IF> and <TMPL_ELSE>.  The file
 written with HTML and these new tags is called a template.  It is
 usually saved separate from your script - possibly even created by
 someone else!  Using this module you fill in the values for the
 variables, loops and branches declared in the template.  This allows
-you to seperate design - the HTML - from the data, which you generate
+you to separate design - the HTML - from the data, which you generate
 in the Perl script.
 
-This module is licenced under the GPL.  See the LICENSE section
+This module is licensed under the GPL.  See the LICENSE section
 below for more details.
 
 =head1 MOTIVATION
@@ -107,6 +109,18 @@ work as planned.
 The "NAME=" in the tag is optional, although for extensibility's sake I
 recommend using it.  Example - "<TMPL_LOOP LOOP_NAME>" is acceptable.
 
+If you're a fanatic about valid HTML and would like your templates
+to conform to valid HTML syntax, you may optionally type template tags
+in the form of HTML comments. This may be of use to HTML authors who
+would like to validate their templates' HTML syntax prior to
+HTML::Template processing, or who use DTD-savvy editing tools.
+
+  <!-- TMPL_VAR NAME=PARAM1 -->
+
+In order to realize a dramatic savings in bandwidth, the standard
+(non-comment) tags will be used throughout the rest of this
+documentation.
+
 =head2 <TMPL_VAR ?ESCAPE=1? NAME="PARAMETER_NAME">
 
 The <TMPL_VAR> tag is very simple.  For each <TMPL_VAR> tag in the
@@ -235,7 +249,7 @@ example of a param call with one nested loop:
 Basically, each <TMPL_LOOP> gets an array reference.  Inside the array
 are any number of hash references.  These hashes contain the
 name=>value pairs for a single pass over the loop template.  It is
-probably in your best interest to build these up programatically, but
+probably in your best interest to build these up programmatically, but
 that is up to you!
 
 Inside a <TMPL_LOOP>, the only variables that are usable are the ones
@@ -249,7 +263,12 @@ This tag includes a template directly into the current template at the
 point where the tag is found.  The included template contents exactly
 as if its contents were physically included in the master template.
 
-NOTE: Currently, each <TMPL_INCLUDE> must be on a seperate line by itself.
+The file specified can be a full path - beginning with a '/'.  If it
+isn't a full path, the path to the enclosing file is tried first.
+After that the path in the environment variable HTML_TEMPLATE_ROOT is
+tried next.  Next, the "path" new() option is consulted.  As a final
+attempt, the filename is passed to open() directly.  See below for
+more information on HTML_TEMPLATE_ROOT and the "path" option to new().
 
 =head2 <TMPL_IF NAME="CONTROL_PARAMETER_NAME"> </TMPL_IF>
 
@@ -280,7 +299,7 @@ that intersecting a <TMPL_IF> and a <TMPL_LOOP> is invalid.
 
 WARNING: Much of the benefit of HTML::Template is in decoupling your
 Perl and HTML.  If you introduce numerous cases where you have
-TMPL_IFs and matching Perl if()s, you will create a maintainance
+TMPL_IFs and matching Perl if()s, you will create a maintenance
 problem in keeping the two synchronized.  I suggest you adopt the
 practice of only using TMPL_IF if you can do so without requiring a
 matching if() in your Perl code.
@@ -311,7 +330,7 @@ Call new() to create a new Template object:
                                       option => 'value' 
                                     );
 
-You must call new() with at least one name => value pair specifing how
+You must call new() with at least one name => value pair specifying how
 to access the template text.  You can use "filename => 'file.tmpl'" to
 specify a filename to be opened as the template.  Alternately you can
 use:
@@ -357,7 +376,8 @@ value of $HTML_TEMPLATE_ROOT.  Example - if the environment variable
 HTML_TEMPLATE_ROOT is set to "/home/sam" and I call
 HTML::Template->new() with filename set to "sam.tmpl", the
 HTML::Template will try to open "/home/sam/sam.tmpl" to access the
-template file.
+template file.  You can also affect the search path for files with the
+"path" option to new() - see below for more information.
 
 You can modify the Template object's behavior with new.  These options
 are available:
@@ -375,7 +395,7 @@ exist in the template body.  Defaults to 1.
 cache - if set to 1 the module will cache in memory the parsed
 templates based on the filename parameter and modification date of the
 file.  This only applies to templates opened with the filename
-parameter specified, not scalarref or arrayref templates.  Cacheing
+parameter specified, not scalarref or arrayref templates.  Caching
 also looks at the modification times of any files included using
 <TMPL_INCLUDE> tags, but again, only if the template is opened with
 filename parameter.  Note that different new() parameter settings do
@@ -388,7 +408,7 @@ performance increase under mod_perl, more if you use large
 =item *
 
 blind_cache - if set to 1 the module behaves exactly as with normal
-cacheing but does not check to see if the file has changed on each
+caching but does not check to see if the file has changed on each
 request.  This option should be used with caution, but could be of use
 on high-load servers.  My tests show blind_cache performing only 1 to
 2 percent faster than cache under mod_perl.
@@ -396,7 +416,7 @@ on high-load servers.  My tests show blind_cache performing only 1 to
 =item *
 
 associate - this option allows you to inherit the parameter values
-from other objects.  The only rwquirement for the other object is that
+from other objects.  The only requirement for the other object is that
 it have a param() method that works like HTML::Template's param().  A
 good candidate would be a CGI.pm query object.  Example:
 
@@ -424,17 +444,28 @@ considered obsolete.
 
 =item *
 
-no_includes - if you know that your template does not have
-TMPL_INCLUDE tags, then you can set no_includes to 1.  This will give
-a small performance gain, since the prepass for include tags can be
-skipped.  Defaults to 0.
+path - you can set this variable with a list of paths to search for
+files specified with the "filename" option to new() and for files
+included with the <TMPL_INCLUDE> tag.  This list is only consulted
+when the filename is relative - i.e. does not begin with a '/'.  The
+HTML_TEMPLATE_ROOT environment variable is always tried first if it
+exists.  In the case of a <TMPL_INCLUDE> file, the path to the
+including file is also tried before path is consulted.
 
+Example:
+
+   my $template = HTML::Template->new( filename => 'file.tmpl',
+                                       path => [ '/path/to/templates',
+                                                 '/alternate/path'
+                                               ]
+                                      );
 =item *
 
 vanguard_compatibility_mode - if set to 1 the module will expect to
 see <TMPL_VAR>s that look like %NAME% instead of the standard syntax.
-If you're not at Vanguard Media trying to use an old format template
-don't worry about this one.  Defaults to 0.
+Also sets die_on_bad_params => 0.  If you're not at Vanguard Media
+trying to use an old format template don't worry about this one.
+Defaults to 0.
 
 =item *
 
@@ -494,8 +525,8 @@ sub new {
                cache_debug => 0,
                die_on_bad_params => 1,
                vanguard_compatibility_mode => 0,
-               no_includes => 0,
                associate => [],
+               path => []
               );
   
   # load in options supplied to new()
@@ -507,6 +538,10 @@ sub new {
   # blind_cache = 1 implies cache = 1
   $options->{blind_cache} and $options->{cache} = 1;
 
+  # vanguard_compatibility_mode implies die_on_bad_params = 1
+  $options->{vanguard_compatibility_mode} and 
+    $options->{die_on_bad_params} = 1;
+
   # handle the "type", "source" parameter format (does anyone use it?)
   if (exists($options->{type})) {
     exists($options->{source}) or die "HTML::Template->new() called with 'type' parameter set, but no 'source'!";
@@ -523,6 +558,11 @@ sub new {
   if (ref($options->{associate}) ne 'ARRAY') {
     $options->{associate} = [ $options->{associate} ];
   }
+
+  # path should be an array if it's not already
+  if (ref($options->{path}) ne 'ARRAY') {
+    $options->{path} = [ $options->{path} ];
+  }
   
   # make sure objects in associate area support param()
   foreach my $object (@{$options->{associate}}) {
@@ -537,7 +577,7 @@ sub new {
   exists($options->{arrayref}) and $source_count++;
   exists($options->{scalarref}) and $source_count++;
   if ($source_count != 1) {
-    die "HTML::Template->new called with multiple (or no) template sources specified!  A valid call to new() has exactly one filename => 'file' OR exactly one scalarRef => \\\$scalar OR exactly one arrayRef = \\\@array";    
+    die "HTML::Template->new called with multiple (or no) template sources specified!  A valid call to new() has exactly one filename => 'file' OR exactly one scalarRef => \\\$scalar OR exactly one arrayRef => \\\@array";    
   }
 
   # initialize data structures
@@ -546,7 +586,7 @@ sub new {
   return $self;
 }
 
-# an internally used new that recieves its parse_stack and param_map as input
+# an internally used new that receives its parse_stack and param_map as input
 sub _new_from_loop {
   my $pkg = shift;
   my $self; { my %hash; $self = bless(\%hash, $pkg); }
@@ -576,7 +616,7 @@ sub _new_from_loop {
   delete($options->{param_map});
   delete($options->{parse_stack});
 
-  return $self;
+ return $self;
 }
 
 # a few shortcuts to new(), of possible use...
@@ -590,20 +630,13 @@ sub new_scalar_ref {
   my $pkg = shift; return $pkg->new('scalarref', @_);
 }
 
-# initilizes all the object data structures.  Also handles global
-# cacheing of template parse data.
+# initializes all the object data structures.  Also handles global
+# caching of template parse data.
 use vars qw( %CACHE );
 sub _init {
   my $self = shift;
   my $options = $self->{options};
 
-  # deal with $ENV{HTML_TEMPLATE_ROOT} stuff
-  if (defined($options->{filename})) {
-    if (!($options->{filename} =~ /^\//) and exists($ENV{HTML_TEMPLATE_ROOT})) {
-      $options->{filename} = join('/', $ENV{HTML_TEMPLATE_ROOT}, split('/', $options->{filename}));
-    }
-  }
-
   # look in the cache to see if we have a cached copy of this
   # template, note modification time in $mtime if we do.
   my $mtime;
@@ -614,7 +647,6 @@ sub _init {
     
     if (!$options->{blind_cache}) {
       # make sure it still exists in the filesystem 
-
       (-r $options->{filename}) or die("HTML::Template : template file $options->{filename} does not exist or is unreadable.");    
 
       # get the modification time
@@ -635,8 +667,7 @@ sub _init {
       }
 
       # if the template has includes, check each included file's mtime
-      # and re-call $self->_init if different.  There's no way,
-      # currently, to just re-read one file...
+      # and re-call $self->_init if different.  
       if (exists($CACHE{$options->{filename}}{included_mtimes})) {
         foreach my $filename (keys %{$CACHE{$options->{filename}}{included_mtimes}}) {
           defined($CACHE{$options->{filename}}{included_mtimes}{$filename}) or
@@ -714,19 +745,52 @@ sub _init_template {
   my $options = $self->{options};
 
   if (exists($options->{filename})) {    
-    # check filename param and try to open
-    (-r $options->{filename}) or die("HTML::Template : template file $options->{filename} does not exist or is unreadable.");
+    my $filename = $options->{filename};
+    my $filepath;
 
-    # open the file
-    open(TEMPLATE, $options->{filename}) or die("Unable to open file $options->{filename}");
+    # find the file and open it
+  SEARCH: {            
+      # first check for a full path
+      if ($filename =~ /^\//) {
+        $filepath = $filename;              
+        die "HTML::Template->new() : Cannot open included file $filename : $!"
+          unless defined(open(TEMPLATE, $filepath));              
+        last SEARCH;                             
+      }            
+      
+      # try HTML_TEMPLATE_ROOT if it exists...
+      
+      if (exists($ENV{HTML_TEMPLATE_ROOT})) {
+        $filepath = join('/', split('/', $ENV{HTML_TEMPLATE_ROOT}), $filename);
+        last SEARCH
+          if (defined(open(TEMPLATE, $filepath)));
+      }
+      
+      # try "path" option list..
+      foreach my $path (@{$options->{path}}) {
+        $filepath = join('/', split('/', $path), $filename);
+        last SEARCH
+          if (defined(open(TEMPLATE, $filepath)));
+      }
 
+      # try even a relative path from the current directory...
+      $filepath = $filename;
+      last SEARCH
+        if (defined(open(TEMPLATE, $filepath)));
+
+      die "HTML::Template->new() : Cannot open template file $filename - file does not exist or is unreadable.";
+    }
+
+    # we'll need this for future reference - to call stat() for example.
+    $options->{filename} = $filepath;
+    
     # read into the array
     my @templateArray = <TEMPLATE>;
     close(TEMPLATE);
-
+    
     # copy in the ref
     $self->{template} = \@templateArray;
-
+    
   } elsif (exists($options->{scalarref})) {
     # split it into an array by line, preserving \n's on all but the
     # last line
@@ -746,46 +810,6 @@ sub _init_template {
     die("HTML::Template : Need to call new with filename, scalarref or arrayref parameter specified.");
   }
 
-  # look for TMPL_INCLUDEs and process them now
-  use vars qw($line);
-  local(*line);
-  if (!($options->{no_includes})) {
-    for (my $line_number = 0; $line_number <= $#{$self->{template}}; $line_number++) {
-      *line = \$self->{template}[$line_number];
-      defined($line) or next;
-
-      if ($line =~ /^<[tT][mM][pP][lL]_[Ii][Nn][Cc][Ll][Uu][Dd][Ee]\s+(?:[nN][aA][mM][eE]\s*=)?\s*"?([\w\/\.]+)"?\s*>$/x) {
-        my $filename = $1;
-
-        # open the file - prefer the INCLUDE to be under the same root
-        # at the template...
-        my @path = split('/', $options->{filename});
-        $path[$#path] = $filename;
-        if (!defined(open(TEMPLATE, join('/', @path)))) {
-          defined(open(TEMPLATE, $filename)) or
-            die "Cannot open included file $filename - tried $filename and " 
-              . join('/', @path);
-        } else {
-          $filename = join('/', @path);
-        }
-      
-        # read into the array
-        my @templateArray = <TEMPLATE>;
-        close(TEMPLATE);
-
-        if (!scalar(@templateArray)) { next; }
-
-        # collect mtimes for included files
-        if ($options->{cache} and !$options->{blind_cache}) {
-          $self->{included_mtimes}{$filename} = (stat($filename))[9];
-        }
-
-        # move the new lines into place.  
-        splice(@{$self->{template}}, $line_number, 1, @templateArray);
-      }
-    }
-  }
-
   return $self;
 }
 
@@ -798,7 +822,7 @@ sub _parse {
   my $self = shift;
   my $options = $self->{options};
   
-  $options->{debug} and print STDERR "### HTML::Template Debug ### In pre_parse:\n";
+  $options->{debug} and print STDERR "### HTML::Template Debug ### In _parse:\n";
   
   # setup the stacks and maps - they're accessed by typeglobs that
   # reference the top of the stack.  They are masked so that a loop
@@ -832,23 +856,26 @@ sub _parse {
   my @loopstack = ();
 
   # loop through lines, filling up pstack
- LINE: for (my $line_number = 0; $line_number <= $#{$self->{template}}; $line_number++) {
+  my $last_line =  $#{$self->{template}};
+ LINE: for (my $line_number = 0; $line_number <= $last_line; $line_number++) {
     next unless defined $self->{template}[$line_number]; 
     my $line = $self->{template}[$line_number];
 
     # handle the old vanguard format
     $options->{vanguard_compatibility_mode} and 
-      $line =~ s/%([\w]+)%/<TMPL_VAR NAME=$1>/g;
+      $line =~ s/%([-\w\/\.+]+)%/<TMPL_VAR NAME=$1>/g;
 
     # make PASSes over the line until nothing is left
   PASS: while(1) {
       last PASS unless defined($line);
       
       # a general regex to match any and all TMPL_* tags
+
       if ($line =~ /
-                    (.*?)
+                    (.*?<)
+                    (?:!--\s*)?
                     (
-                      <\/?[tT][mM][pP][lL]_
+                      \/?[tT][mM][pP][lL]_
                       (?:
                          (?:[Vv][Aa][Rr])
                          |
@@ -857,6 +884,8 @@ sub _parse {
                          (?:[Ii][Ff])
                          |
                          (?:[Ee][Ll][Ss][Ee])
+                         |
+                         (?:[Ii][Nn][Cc][Ll][Uu][Dd][Ee])
                       )
                     )
                     (?:
@@ -878,20 +907,21 @@ sub _parse {
                     \s*
                     "?
                     (
-                     \w+
+                     [-\w\/\.+]+
                     )?
                     "?
                     \s*
-                    >
+                    (?:--)?>
                     (.*)
                    /sgx) {
         my $pre = $1; # what comes before
-        my $which = $2; # which tag is it
+        chop $pre; # remove trailing <
+        my $which = uc($2); # which tag is it
         my $escape = $3;
         my $name = lc $4; # what name for the tag?  undef for a /tag
         my $post = $5; # what comes after on the line
 
-        if ($which =~ /^<[tT][mM][pP][lL]_[Vv][Aa][Rr]/) {
+        if ($which eq 'TMPL_VAR') {
           $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : parsed VAR $name\n";
           
           # if we already have this var, then simply link to the existing
@@ -925,8 +955,8 @@ sub _parse {
           $line = $post;         
           next PASS;          
         
-        } elsif ($which =~ /^<[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp]/) {
-          # we've got a loop start
+        } elsif ($which eq 'TMPL_LOOP') {
+         # we've got a loop start
           $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : LOOP $name start\n";
 
           defined $escape and 
@@ -972,7 +1002,7 @@ sub _parse {
           $line = $post;
           next PASS;
 
-        } elsif ($which =~ /^<\/[tT][mM][pP][lL]_[Ll][Oo][Oo][Pp]/) {
+        } elsif ($which eq '/TMPL_LOOP') {
           $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : LOOP end\n";
 
           defined $escape and 
@@ -998,7 +1028,7 @@ sub _parse {
           my $parse_stack = pop(@pstacks);
           *pstack = $pstacks[$#pstacks];
           
-          scalar(@ifstack) and die "HTML::Template->new() : Dangleing <TMPL_IF> in loop ending at $line_number!";
+          scalar(@ifstack) and die "HTML::Template->new() : Dangling <TMPL_IF> in loop ending at $line_number!";
           pop(@ifstacks);
           *ifstack = $ifstacks[$#ifstack];
           
@@ -1018,7 +1048,7 @@ sub _parse {
           $line = $post;
           next PASS;
           
-        } elsif ($which =~ /^<[Tt][Mm][Pp][Ll]_[Ii][Ff]/) {
+        } elsif ($which eq 'TMPL_IF') {
           $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : IF $name start\n";
 
           defined($name) or die "HTML::Template->new() : found TMPL_IF with no name at $line_number!";
@@ -1056,14 +1086,14 @@ sub _parse {
           $line = $post;
           next PASS;
 
-        } elsif ($which =~ /^<\/[Tt][Mm][Pp][Ll]_[Ii][Ff]/x) {
+        } elsif ($which eq '/TMPL_IF') {
           $options->{debug} and print STDERR "### HTML::Template Debug ###$line_number : IF end\n";
 
           defined $escape and 
             die "HTML::Template->new() : Found ESCAPE option in a /TMPL_IF tag at line $line_number.  ESCAPE is only valid for TMPL_VARs.";
 
           my $if = pop(@ifstack);
-          die "HTML::Template->new() : found </TMPL_IF|ELSE> with no matching <TMPL_IF|ELSE> at $line_number!" unless defined $if;
+          die "HTML::Template->new() : found </TMPL_IF> with no matching <TMPL_IF|ELSE> at $line_number!" unless defined $if;
 
           # push text coming before the tag onto the pstack,
           # concatenating with preceding text if possible.
@@ -1083,7 +1113,7 @@ sub _parse {
           $line = $post;
           next PASS;
 
-        } elsif ($which =~ /^<[Tt][Mm][Pp][Ll]_[Ee][Ll][Ss][Ee]/) {
+        } elsif ($which eq 'TMPL_ELSE') {
           $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : ELSE\n";
 
           defined $escape and 
@@ -1106,7 +1136,7 @@ sub _parse {
 
           # connect the matching to this "address" - thus the if,
           # failing jumps to the ELSE address.  The else then gets
-          # elaborated, and of course succedes.  On the other hand, if
+          # elaborated, and of course succeeds.  On the other hand, if
           # the IF fails and falls though, output will reach the else
           # and jump to the /if address.
           $if->[HTML::Template::IF::JUMP_ADDRESS] = $#pstack;
@@ -1114,6 +1144,93 @@ sub _parse {
           $line = $post;
           next PASS;
 
+        } elsif ($which eq 'TMPL_INCLUDE') {
+          # handle TMPL_INCLUDEs
+          $options->{debug} and print STDERR "### HTML::Template Debug ### $line_number : INCLUDE $name \n";
+
+          defined $name or die "HTML::Template->new() : found TMPL_INCLUDE with no name at $line_number!";
+
+          defined $escape and 
+            die "HTML::Template->new() : Found ESCAPE option in a TMPL_INCLUDE tag at line $line_number.  ESCAPE is only valid for TMPL_VARs.";
+
+          # push text coming before the tag onto the pstack,
+          # concatenating with preceding text if possible.
+          if (ref($pstack[$#pstack]) eq 'SCALAR') {
+            ${$pstack[$#pstack]} .= $pre;
+          } else {
+            push(@pstack, \$pre);
+          }
+
+          my $filename = $name;
+          my $filepath;
+
+          # look for the included file...
+        SEARCH: {            
+            # first check for a full path
+            if ($filename =~ /^\//) {
+              $filepath = $filename;              
+              die "HTML::Template->new() : Cannot open included file $filename : $!"
+                unless defined(open(TEMPLATE, $filepath));              
+              last SEARCH;                             
+            }            
+            
+            # try the path of the enclosing template if it has one
+            if (exists($options->{filename})) {
+              my @path = split('/', $options->{filename});
+              $path[$#path] = $filename;
+              $filepath = join('/', @path);
+              last SEARCH
+                if (defined(open(TEMPLATE, $filepath)));
+            }
+
+            # try HTML_TEMPLATE_ROOT if it exists...
+            if (exists($ENV{HTML_TEMPLATE_ROOT})) {
+              $filepath = join('/', split('/', $ENV{HTML_TEMPLATE_ROOT}), $filename);
+              last SEARCH
+                if (defined(open(TEMPLATE, $filepath)));
+            }
+            
+            # try "path" option list..
+            foreach my $path (@{$options->{path}}) {
+              $filepath = join('/', split('/', $path), $filename);
+              last SEARCH
+                if (defined(open(TEMPLATE, $filepath)));
+            }
+
+            # try even a relative path from the current directory...
+            $filepath = $filename;
+            last SEARCH
+              if (defined(open(TEMPLATE, $filepath)));
+
+            die "HTML::Template->new() : Cannot open included file $filename";
+          }
+      
+          # read into the array
+          my @templateArray = <TEMPLATE>;
+          close(TEMPLATE);
+          
+          $line = $post, next PASS 
+            unless (scalar(@templateArray));
+
+          # collect mtimes for included files
+          if ($options->{cache} and !$options->{blind_cache}) {
+            $self->{included_mtimes}{$filepath} = (stat($filepath))[9];
+          }
+
+          # stick the remains of this line onto the bottom of the
+          # included text.
+          push(@templateArray, $post);
+          
+          # move the new lines into place.  
+          splice(@{$self->{template}}, $line_number, 1, @templateArray);
+          
+          # recalculate stopping point
+          $last_line = $#{$self->{template}};
+          
+          # start in on the first line of the included text - nothing
+          # else to do on this line.
+          $line = $self->{template}[$line_number];
+          next PASS;
         } else {
           # zuh!?
           die "HTML::Template->new() : Unknown or unmatched TMPL construct on line $line_number";
@@ -1149,7 +1266,7 @@ sub _parse {
 
 =head2 param
 
-param() can be called in three ways
+param() can be called in a number of ways
 
 
 1) To return a list of parameters in the template : 
@@ -1176,21 +1293,69 @@ param() can be called in three ways
                    ]
                   );
 
+4) To set the value of a a number of parameters :
+
+     # For simple TMPL_VARs:
+     $self->param(PARAM => 'value', 
+                  PARAM2 => 'value'
+                 );
+
+      # And with some TMPL_LOOPs:
+      $self->param(PARAM => 'value', 
+                   PARAM2 => 'value',
+                   LOOP_PARAM => 
+                   [ 
+                    { PARAM => VALUE_FOR_FIRST_PASS, ... }, 
+                    { PARAM => VALUE_FOR_SECOND_PASS, ... } 
+                    ...
+                   ],
+                   ANOTHER_LOOP_PARAM => 
+                   [ 
+                    { PARAM => VALUE_FOR_FIRST_PASS, ... }, 
+                    { PARAM => VALUE_FOR_SECOND_PASS, ... } 
+                    ...
+                   ]
+                  );
+
+5) To set the value of a a number of parameters using a hash-ref :
+
+      $self->param(
+                   { 
+                      PARAM => 'value', 
+                      PARAM2 => 'value',
+                      LOOP_PARAM => 
+                      [ 
+                        { PARAM => VALUE_FOR_FIRST_PASS, ... }, 
+                        { PARAM => VALUE_FOR_SECOND_PASS, ... } 
+                        ...
+                      ],
+                      ANOTHER_LOOP_PARAM => 
+                      [ 
+                        { PARAM => VALUE_FOR_FIRST_PASS, ... }, 
+                        { PARAM => VALUE_FOR_SECOND_PASS, ... } 
+                        ...
+                      ]
+                    }
+                   );
+
 =cut
 
 
 sub param {
-  my ($self, $param, $value) = @_;
+  my $self = shift;
   my $options = $self->{options};
   my $param_map = $self->{param_map};
-  my $type;
 
-  if (!defined($param)) {
-    # return a list of parameters in this template
-    return keys(%{$param_map});
+  # the no-parameter case - return list of parameters in the template.
+  return keys(%{$param_map}) unless scalar(@_);
+  
+  my $first = shift;
+  my $type = ref $first;
 
-  } elsif (!defined($value)) {
-    $param = lc($param);
+  # the one-parameter case - could be a parameter value request or a
+  # hash-ref.
+  if (!scalar(@_) and !length($type)) {
+    my $param = lc $first;
     
     # check for parameter existence 
     $options->{die_on_bad_params} and !exists($param_map->{$param}) and
@@ -1205,29 +1370,45 @@ sub param {
     ($type eq 'HTML::Template::LOOP') and
       return $param_map->{$param}[HTML::Template::LOOP::PARAM_SET];
     die "Unknown param type $type!";
+  } 
+
+  if (!scalar(@_)) {
+    die "HTML::Template->param() : Single reference arg to param() must be a hash-ref!  You gave me a $type." 
+      unless $type eq 'HASH';  
+    push(@_, %{$first});
   } else {
-    $param = lc($param);
+    unshift(@_, $first);
+
+  }
+  
+  die "HTML::Template->param() : You gave me an odd number of parameters to param()!  Read the docs, man." 
+    unless ((@_ % 2) == 0);
+
+  for (my $x = 0; $x <= $#_; $x += 2) {
+    my $param = lc $_[$x];
+    my $value = $_[($x + 1)];
+
     # check that this param exists in the template
     $options->{die_on_bad_params} and !exists($param_map->{$param}) and
       die("HTML::Template : Attempt to set nonexistent parameter $param : (die_on_bad_params => 1)");
     
-    return unless (exists($param_map->{$param}) and
-                   defined($param_map->{$param}));
+    # if we're not going to do from bad param names, we need to ignore
+    # them...
+    next unless (exists($param_map->{$param}));
     
     # copy in contents of ARRAY refs to prevent confusion - 
     # thanks Richard!
-    $type = ref($param_map->{$param});
     if ( ref($value) eq 'ARRAY' ) {
-      ($type eq 'HTML::Template::LOOP') or
+      (ref($param_map->{$param}) eq 'HTML::Template::LOOP') or
         die "HTML::Template::param() : attempt to set parameter $param with an array ref - parameter is not a TMPL_LOOP!";
       $param_map->{$param}[HTML::Template::LOOP::PARAM_SET] = [@{$value}];
     } else {
-      ($type eq 'HTML::Template::VAR') or
+      (ref($param_map->{$param}) eq 'HTML::Template::VAR') or
         die "HTML::Template::param() : attempt to set parameter $param with a scalar - parameter is not a TMPL_VAR!";
       ${$param_map->{$param}} = $value;
     }
-    return 1;
   }
+  return 1;
 }
 
 =head2 clear_params()
@@ -1266,7 +1447,7 @@ output() returns the final result of the template.  In most situations you'll wa
 
    print $template->output();
 
-When output is called each occurance of <TMPL_VAR NAME=name> is
+When output is called each occurrence of <TMPL_VAR NAME=name> is
 replaced with the value assigned to "name" via param().  If a named
 parameter is unset it is simply replaced with ''.  <TMPL_LOOPS> are
 evaluated once per parameter set, accumlating output on each pass.
@@ -1278,12 +1459,11 @@ important for the internal implementation of loops.
 =cut
 
 
-
 sub output {
   my $self = shift;
   my $options = $self->{options};
 
-  $options->{debug} and print STDERR "### HTML::Template Debug ### In output $self\n";
+  $options->{debug} and print STDERR "### HTML::Template Debug ### In output\n";
 
   # want a stack dump?
   if ($options->{debug_stack}) {
@@ -1313,7 +1493,8 @@ sub output {
   *parse_stack = $self->{parse_stack};
   my $result = '';
   my $type;
-  for (my $x = 0; $x <= $#parse_stack; $x++) {
+  my $parse_stack_length = $#parse_stack;
+  for (my $x = 0; $x <= $parse_stack_length; $x++) {
     *line = \$parse_stack[$x];
     $type = ref($line);
 
@@ -1401,9 +1582,7 @@ sub output {
   
   my $result = '';
   foreach my $value_set (@$value_sets_array) {
-    foreach my $name (keys %$value_set) {
-      $template->param($name, $value_set->{$name});
-    }
+    $template->param($value_set);
     $result .= $template->output;
     $template->clear_params;
   }
@@ -1537,6 +1716,10 @@ instead getting a speed increase, you'll double your memory usage.  To
 find out if this is happening set cache_debug => 1 in your application
 code and look for "CACHE MISS" messages in the logs.
 
+=head2 What characters are allowed in TMPL_* NAMEs?
+
+Numbers, letters, '.', '/', '+', '-' and '_'.
+
 =head1 BUGS
 
 I am aware of no bugs - if you find one, join the mailing list and
@@ -1572,6 +1755,9 @@ provided by:
    David Glasser
    Peter Marelas
    James William Carlson
+   Frank D. Cringle
+   Winfried Koenig
+   Matthew Wickline
 
 Thanks!
 
